from django.conf import settings as django_settings

DEFAULT_SETTINGS = {
    "WHISPERER_REQUEST_TIMEOUT": 10,
}


class Settings(object):
    def __getattr__(self, item):
        if item in DEFAULT_SETTINGS:
            return getattr(django_settings, item, DEFAULT_SETTINGS[item])
        raise AttributeError


settings = Settings()



from django.core import serializers

from whisperer.exceptions import AlreadyRegisteredEvent, ImproperlyConfigured


class WhispererEvent(object):
    serializer_class = None
    event_type = None

    def _serialize(self, instance):
        objects = [instance] if not isinstance(instance, list) else instance
        return serializers.serialize('json', objects)

    def serialize(self, instance):
        if isinstance(instance, dict):
            return instance

        if self.serializer_class is None:
            return self._serialize(instance)

        many = isinstance(instance, list)
        return self.serializer_class(instance, many=many).data


class EventRegistry(object):
    def __init__(self):
        self._registry = {}

    def __contains__(self, event_type):
        return event_type in self._registry

    def __getitem__(self, event_type):
        return self._registry[event_type]

    @property
    def event_types(self):
        return self._registry.keys()

    @property
    def event_type_choices(self):
        event_types = self.event_types
        return zip(event_types, event_types)

    def register(self, *events):
        for event in events:
            if not issubclass(event, WhispererEvent):
                raise ImproperlyConfigured()

            if event.event_type in self._registry:
                raise AlreadyRegisteredEvent(params=(event.event_type,))

            self._registry[event.event_type] = event


registry = EventRegistry()



undefined = {"code": "undefined", "en": "undefined exception code"}

webhook_100_1 = {"code": "webhook_100_1", "en": "Webhook does not exist"}

webhook_100_2 = {"code": "webhook_100_2", "en": "Webhook already registered"}

event_100_1 = {"code": "event_100_1", "en": "WebhookEvent does not exist"}

event_100_2 = {"code": "event_100_2", "en": "Unknown event type"}

event_100_3 = {"code": "event_100_3", "en": "Event already delivered"}

event_100_4 = {
    "code": "event_100_4",
    "en": "Events must be inherited from WhispererEvent",
}

event_100_5 = {"code": "event_100_5", "en": "{event_type} is already registered"}



try:
    from django.urls import include, path
except ImportError:
    from django.conf.urls import include, url as path

from rest_framework.routers import DefaultRouter

from whisperer.resources.views import WebhookEventViewSet, WebhookViewSet

router = DefaultRouter()
router.register(r'hooks', WebhookViewSet)
router.register(r'hook_events', WebhookEventViewSet)

app_name = 'whisperer'


urlpatterns = [path('', include(router.urls))]



import sys

from whisperer.models import Webhook, WebhookEvent

if sys.version_info.major == 3:
    from django_filters import FilterSet, filters

    class WebhookFilterPY3(FilterSet):
        target_url = filters.CharFilter(field_name="target_url", lookup_expr="iexact")
        event_type = filters.CharFilter(field_name="event_type", lookup_expr="iexact")
        is_active = filters.BooleanFilter(field_name="is_active")

        class Meta:
            model = Webhook
            fields = ('target_url', 'is_active', 'event_type')

    class WebhookEventFilterPY3(FilterSet):
        event_type = filters.CharFilter(
            field_name='webhook__event_type', lookup_expr='iexact'
        )
        created_date = filters.DateFilter(
            field_name='created_date', lookup_expr='__all__'
        )
        delivered = filters.BooleanFilter(field_name="delivered")

        class Meta:
            model = WebhookEvent
            fields = (
                'uuid',
                'event_type',
                'response_http_status',
                'delivered',
                'created_date',
            )

    WebhookFilter = WebhookFilterPY3
    WebhookEventFilter = WebhookEventFilterPY3
else:
    import rest_framework_filters as PY2Filters

    class WebhookFilterPY2(PY2Filters.FilterSet):
        target_url = PY2Filters.CharFilter(name="target_url", lookup_expr="iexact")
        event_type = PY2Filters.CharFilter(name="event_type", lookup_expr="iexact")
        is_active = PY2Filters.BooleanFilter(name='is_active')

        class Meta:
            model = Webhook
            fields = ('target_url', 'is_active', 'event_type')

    class WebhookEventFilterPY2(PY2Filters.FilterSet):
        event_type = PY2Filters.CharFilter(
            name='webhook__event_type', lookup_expr='iexact'
        )
        created_date = PY2Filters.DateFilter(name='created_date', lookup_expr='__all__')
        delivered = PY2Filters.BooleanFilter(name='delivered')

        class Meta:
            model = WebhookEvent
            fields = (
                'uuid',
                'event_type',
                'response_http_status',
                'delivered',
                'created_date',
            )

    WebhookFilter = WebhookFilterPY2()
    WebhookEventFilter = WebhookEventFilterPY2()



from django.db import models
from django.utils.translation import ugettext_lazy as _


class Customer(models.Model):
    email = models.EmailField(_(u'Email'))
    first_name = models.CharField(_("First name"), max_length=255, blank=True)
    last_name = models.CharField(_("Last name"), max_length=255, blank=True)
    phone_number = models.CharField(
        _("Phone number"), null=True, blank=True, max_length=128
    )

    is_active = models.BooleanField(_(u'Is active'), default=True)


class Address(models.Model):
    first_name = models.CharField(_("First name"), max_length=255, blank=True)
    last_name = models.CharField(_("Last name"), max_length=255, blank=True)
    line = models.CharField(verbose_name=_(u'Line of address'), max_length=255)
    is_active = models.BooleanField(default=True)


class Order(models.Model):
    customer = models.ForeignKey(Customer, null=True, on_delete=models.SET_NULL)
    address = models.ForeignKey(Address, null=True, on_delete=models.SET_NULL)
    number = models.CharField(_("Order number"), max_length=128, db_index=True)
    date_placed = models.DateTimeField(_('Date Placed'), db_index=True)
    amount = models.DecimalField(_("Order amount"), decimal_places=2, max_digits=12)
    discount_amount = models.DecimalField(
        _("Discount"), decimal_places=2, max_digits=12, default=0
    )

    shipping_amount = models.DecimalField(
        _("Shipping charge"), decimal_places=2, max_digits=12, default=0
    )



import json
import logging

import requests
from celery import current_app
from django.apps import apps
from django.db import models
from django.db.models import F
from django.db.models.functions import Coalesce
from django.utils import timezone
from django.utils.module_loading import import_string
from rest_framework.utils.encoders import JSONEncoder

from whisperer.client import WebhookClient
from whisperer.events import registry
from whisperer.exceptions import (
    EventAlreadyDelivered,
    UnknownEventType,
    WebhookEventDoesNotExist,
)
from whisperer.utils import LockTask

TASK_RETRY_COUNT = 10
MAX_RETRY_COUNT = 18
logger = logging.getLogger(__name__)


def get_natural_key(instance):
    from django.contrib.contenttypes.models import ContentType

    content_type = ContentType.objects.get_for_model(instance)
    return content_type.natural_key()


@current_app.task(bind=True, max_retries=TASK_RETRY_COUNT, base=LockTask)
def deliver_event_task(
    self,
    hook_id,
    event_type,
    event_uuid=None,
    instance=None,
    app_label=None,
    model_name=None,
    pk=None,
    retry=True,
    **kwargs
):
    from whisperer.models import Webhook

    hook = Webhook.objects.get(pk=hook_id)
    if not event_uuid and not instance and app_label and model_name and pk:
        model_class = apps.get_model(app_label, model_name)
        instance = model_class.objects.get(pk=pk)

    webhook_event, response = _deliver_event(
        hook, instance, event_type, event_uuid=event_uuid
    )

    if not response.ok:
        self.request.kwargs['event_uuid'] = webhook_event.uuid
        webhook_event.refresh_from_db(fields=['retry_count'])
        webhook_event_retry_count = webhook_event.retry_count or 1
        if (
            self.request.retries >= TASK_RETRY_COUNT
            or webhook_event_retry_count >= MAX_RETRY_COUNT
            or not retry
        ):
            return

        self.retry(countdown=hook.countdown.get_value(self.request.retries))


def _deliver_event(hook, instance, event_type, event_uuid=None, force=False):
    from django.contrib.contenttypes.models import ContentType

    from whisperer.models import WebhookEvent

    if event_type not in registry:
        raise UnknownEventType()

    if event_uuid:
        try:
            webhook_event = WebhookEvent.objects.get(uuid=event_uuid)
            payload = webhook_event.request_payload
            if webhook_event.delivered and not force:
                raise EventAlreadyDelivered()
        except WebhookEvent.DoesNotExist:
            raise WebhookEventDoesNotExist()
    else:
        webhook_event = WebhookEvent(webhook=hook, retry_count=0)
        event_class = registry[event_type]
        event = event_class()
        serialize_instance = event.serialize(instance)
        payload = {
            'event': {'type': event_type, 'uuid': webhook_event.uuid.hex},
            'payload': serialize_instance,
        }

    request_datetime = timezone.now()
    response = requests.Response()
    try:
        client = WebhookClient(event_type=event_type, payload=payload)
        response = client.send_payload(
            target_url=hook.target_url,
            payload=payload,
            secret_key=hook.secret_key,
            additional_headers=hook.additional_headers,
            auth_config=hook.config.get('auth'),
        )
    except requests.exceptions.RequestException as exc:
        response.status_code = (exc.response and exc.response.status_code) or 500
        response._content = exc
    except Exception as exc:
        response._content = ''
        response.status_code = 500
        logger.exception(exc)
    finally:
        webhook_event.request_payload = json.loads(json.dumps(payload, cls=JSONEncoder))
        webhook_event.response_content = response.content
        webhook_event.response_http_status = response.status_code
        if isinstance(instance, (models.Model, models.base.ModelBase)):
            webhook_event.object_id = instance.pk
            webhook_event.content_object = instance
            webhook_event.content_type = ContentType.objects.get_for_model(
                instance._meta.model
            )
        if 200 <= response.status_code < 300:
            webhook_event.delivered = True
        else:
            webhook_event.delivered = False
        webhook_event.request_datetimes.insert(0, request_datetime)
        webhook_event.retry_count = (
            Coalesce(F('retry_count') + 1, len(webhook_event.request_datetimes))
            if webhook_event.pk
            else 1
        )
        webhook_event.save()

    if hook.callback:
        callback_function = import_string(hook.callback)
        callback_function(response, event_type, instance, payload)

    return webhook_event, response


def deliver_event(instance, event_type, async_=True, event_uuid=None):
    from whisperer.models import Webhook

    hooks = Webhook.objects.filter(event_type=event_type, is_active=True)
    for hook in hooks:
        if not async_:
            _deliver_event(hook, instance, event_type, event_uuid)
            continue
        if isinstance(instance, (models.Model, models.base.ModelBase)):
            app_label, model_name = get_natural_key(instance)
            deliver_event_task.delay(
                hook_id=hook.pk,
                event_type=event_type,
                app_label=app_label,
                model_name=model_name,
                pk=instance.pk,
                event_uuid=event_uuid,
            )
        elif isinstance(instance, dict):
            deliver_event_task.delay(
                hook_id=hook.pk,
                event_type=event_type,
                instance=instance,
                event_uuid=event_uuid,
            )
        else:
            raise NotImplementedError()


@current_app.task()
def undelivered_event_scanner():
    from whisperer.models import WebhookEvent

    undelivered_events = WebhookEvent.objects.filter(
        retry_count__gte=TASK_RETRY_COUNT + 1,
        retry_count__lte=MAX_RETRY_COUNT,
        delivered=False,
    ).all()

    for undelivered_event in undelivered_events:
        if undelivered_event.is_retry_allowed:
            deliver_event_task.delay(
                hook_id=undelivered_event.webhook_id,
                event_type=undelivered_event.webhook.event_type,
                event_uuid=undelivered_event.uuid,
                retry=False,
            )



import hashlib
import hmac
import json

import requests
from requests.auth import HTTPBasicAuth
from rest_framework.utils.encoders import JSONEncoder

from whisperer.conf import settings


class WebhookClient(object):
    def __init__(self, event_type, payload):
        self.event_type = event_type
        self.payload = payload
        self.headers = {
            'Content-Type': 'application/json',
            'X-Whisperer-Event': self.event_type,
        }

    def sign(self, secret_key, payload):
        signature = hmac.new(
            secret_key.encode('utf-8'),
            payload.encode('utf-8'),
            digestmod=hashlib.sha256,
        )
        self.headers['X-Whisperer-Signature'] = 'sha256={}'.format(
            signature.hexdigest()
        )

    def send_payload(
        self,
        target_url,
        payload,
        secret_key=None,
        additional_headers=None,
        auth_config=None,
        *args,
        **kwargs
    ):
        payload = json.dumps(payload, cls=JSONEncoder)
        auth = None
        if secret_key:
            self.sign(secret_key, payload)
        if auth_config:
            auth = self._get_auth(auth_config)
        if additional_headers:
            self.headers.update(additional_headers)
        response = requests.post(
            url=target_url,
            data=payload,
            headers=self.headers,
            timeout=settings.WHISPERER_REQUEST_TIMEOUT,
            auth=auth,
        )
        return response

    def _get_auth(self, auth_config):
        if auth_config.get('auth_type') == 'basic':
            username = auth_config.get('username')
            password = auth_config.get('password')
            return HTTPBasicAuth(username, password)



import logging
from datetime import datetime

from celery import current_app
from django.core.cache import cache
from django.utils.encoding import force_str

logger = logging.getLogger(__name__)


class LockTask(current_app.Task):
    """this abstract class ensures the same tasks run only once at a time"""

    abstract = True

    def __init__(self, *args, **kwargs):
        super(LockTask, self).__init__(*args, **kwargs)

    @staticmethod
    def is_exists_cache_key(lock_cache_key):
        return True if cache.get(lock_cache_key) else False

    def generate_lock_cache_key(self, *args, **kwargs):
        args_key = [force_str(arg) for arg in args]
        kwargs_key = [
            '{}_{}'.format(k, force_str(v)) for k, v in sorted(kwargs.items())
        ]
        return '_'.join([self.name] + args_key + kwargs_key)

    def __call__(self, *args, **kwargs):
        """check task"""
        lock_cache_key = self.generate_lock_cache_key(*args, **kwargs)
        lock_time = datetime.now().isoformat()
        if not self.is_exists_cache_key(lock_cache_key):
            cache.set(lock_cache_key, lock_time, timeout=2 ** self.request.retries)
            return self.run(*args, **kwargs)
        else:
            logger.info("Task %s is already running.." % self.name)


class Registry(object):
    def __init__(self):
        self._registry = {}

    def __contains__(self, key):
        return key in self._registry

    def __getitem__(self, key):
        return self._registry[key]

    def register(self, key):
        def decorator(cls):
            self._registry[key] = cls
            return cls

        return decorator



import random
import uuid
from datetime import timedelta

import mock
import requests
import requests_mock
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.test import TestCase, override_settings
from django.utils import timezone
from model_mommy import mommy

from whisperer.countdown import (
    ExponentialRetryCountdown,
    FixedRetryCountdown,
    LinearRetryCountdown,
    RandomRetryCountdown,
)
from whisperer.exceptions import WebhookAlreadyRegistered
from whisperer.models import Webhook, WebhookEvent
from whisperer.resources.serializers import WebhookSerializer
from whisperer.services import WebhookService
from whisperer.tasks import (
    MAX_RETRY_COUNT,
    TASK_RETRY_COUNT,
    deliver_event,
    deliver_event_task,
    undelivered_event_scanner,
)
from whisperer.tests.models import Address, Customer, Order
from whisperer.utils import LockTask


class WebhookTestCase(TestCase):
    def setUp(self):
        self.user = mommy.make(User, username="test_user")
        self.service = WebhookService()

    def test_register_webhook(self):
        webhook = mommy.prepare(
            Webhook,
            user=self.user,
            target_url="http://example.com/order_create",
            secret_key="secret",
            event_type="order-created",
            retry_countdown_config={"choice": "exponential", "kwargs": {"base": 2}},
        )

        serializer = WebhookSerializer(webhook)
        serializer = WebhookSerializer(data=serializer.data)
        self.assertTrue(serializer.is_valid(raise_exception=True))
        webhook = self.service.register_webhook(
            user=self.user, **serializer.validated_data
        )
        self.assertEqual(webhook.event_type, "order-created")

        with self.assertRaises(WebhookAlreadyRegistered):
            self.service.register_webhook(user=self.user, **serializer.validated_data)

    def test_update_webhook(self):
        webhook = mommy.prepare(
            Webhook,
            user=self.user,
            target_url="http://example2.com/order_create",
            secret_key="secret",
            event_type="order-created",
            retry_countdown_config={"choice": "exponential", "kwargs": {"base": 2}},
        )

        serializer = WebhookSerializer(webhook)
        serializer = WebhookSerializer(data=serializer.data)
        self.assertTrue(serializer.is_valid(raise_exception=True))
        webhook_created = self.service.register_webhook(
            user=self.user, **serializer.validated_data
        )

        update_data = {
            "target_url": "http://example3.com/order_update",
            "event_type": "order-update",
            "retry_countdown_config": {
                "choice": "linear",
                "kwargs": {"base": 1 * 60, "limit": 5 * 60},
            },
        }
        webhook_updated = self.service.update_webhook(
            webhook_created, self.user, **update_data
        )

        self.assertNotEqual(webhook.target_url, webhook_updated.target_url)
        self.assertNotEqual(webhook.event_type, webhook_updated.event_type)

        self.assertEqual(webhook_updated.target_url, update_data["target_url"])
        self.assertEqual(webhook_updated.event_type, "order-update")

    def test_update_webhook_with_already_registered_one(self):
        webhook = mommy.make(
            Webhook,
            user=self.user,
            target_url="http://example2.com/order_create",
            secret_key="secret",
            event_type="order-created",
            retry_countdown_config={"choice": "exponential", "kwargs": {"base": 2}},
        )

        mommy.make(
            Webhook,
            user=self.user,
            target_url="http://example2.com/order_create",
            secret_key="secret",
            event_type="order-update",
            retry_countdown_config={"choice": "exponential", "kwargs": {"base": 2}},
        )

        update_data = {"target_url": webhook.target_url, "event_type": "order-update"}

        with self.assertRaises(WebhookAlreadyRegistered):
            self.service.update_webhook(webhook, self.user, **update_data)

        del update_data["target_url"]

        with self.assertRaises(WebhookAlreadyRegistered):
            self.service.update_webhook(webhook, self.user, **update_data)

        update_data = {
            "retry_countdown_config": {
                "choice": "linear",
                "kwargs": {"base": 1 * 60, "limit": 10 * 60},
            },
        }

        webhook_updated = self.service.update_webhook(webhook, self.user, **update_data)
        self.assertEqual(
            webhook_updated.retry_countdown_config["kwargs"],
            update_data["retry_countdown_config"]["kwargs"],
        )

    def test_delete_webhook(self):
        webhook = mommy.make(
            Webhook,
            is_active=True,
            retry_countdown_config={"choice": "exponential", "kwargs": {"base": 2}},
        )
        self.service.delete_webhook(webhook)
        webhook.refresh_from_db()
        self.assertFalse(webhook.is_active)


@override_settings(CELERY_ALWAYS_EAGER=True)
class WhispererEventTestCase(TestCase):
    def setUp(self):
        self.target_url = "http://example.com/order_create"
        self.user = mommy.make(User, username="test_user")

        self.webhook = mommy.make(
            Webhook,
            user=self.user,
            target_url=self.target_url,
            secret_key="secret",
            event_type="order-created",
            retry_countdown_config={"choice": "exponential", "kwargs": {"base": 2}},
        )
        self.target_url2 = "http://example.com/auth_order_update"
        self.webhook2 = mommy.make(
            Webhook,
            user=self.user,
            target_url=self.target_url2,
            secret_key="secret",
            event_type="order-updated",
            retry_countdown_config={"choice": "exponential", "kwargs": {"base": 2}},
            config={
                "auth": {
                    "username": "username",
                    "password": "123",
                    "auth_type": "basic",
                }
            },
        )
        self.customer = mommy.make(Customer)
        self.address = mommy.make(Address)
        self.order = mommy.prepare(
            Order,
            customer=self.customer,
            address=self.address,
            number="1",
            amount="1",
            discount_amount="1",
            shipping_amount="1",
        )

    def test_deliver_event(self):
        with requests_mock.Mocker() as mock:
            mock.register_uri(
                "POST", self.target_url, text="Order Created", status_code=200
            )
            self.order.save()
            webhook_events = WebhookEvent.objects.all()
            self.assertEqual(len(webhook_events), 1)
            self.assertTrue(webhook_events[0].delivered)

    def test_deliver_event_sync_with_event_uuid(self):
        webhookevent = mommy.make(
            WebhookEvent,
            modified_date=timezone.now(),
            created_date=timezone.now(),
            request_payload={},
            retry_count=TASK_RETRY_COUNT,
            delivered=False,
            webhook=self.webhook,
        )
        with requests_mock.Mocker() as mock:
            mock.register_uri(
                "POST", self.webhook.target_url, text="Order Created", status_code=200
            )
            deliver_event(
                self.order, "order-created", async_=False, event_uuid=webhookevent.uuid
            )
            webhookevent.refresh_from_db()
            self.assertTrue(webhookevent.delivered)

    def test_deliver_event_async_with_event_uuid(self):
        webhookevent = mommy.make(
            WebhookEvent,
            modified_date=timezone.now(),
            created_date=timezone.now(),
            request_payload={},
            retry_count=TASK_RETRY_COUNT,
            delivered=False,
            webhook=self.webhook,
        )
        with requests_mock.Mocker() as mock:
            mock.register_uri(
                "POST", self.webhook.target_url, text="Order Created", status_code=200
            )
            deliver_event(self.order, "order-created", event_uuid=webhookevent.uuid)
            webhookevent.refresh_from_db()
            self.assertTrue(webhookevent.delivered)

    @mock.patch("requests.post")
    def test_http_error(self, post_mock):
        post_mock.side_effect = requests.exceptions.HTTPError("HTTPError")
        self.order.save()
        webhook_events = WebhookEvent.objects.all()
        self.assertEqual(len(webhook_events), 1)
        self.assertFalse(webhook_events[0].delivered)
        self.assertEqual(webhook_events[0].response_http_status, 500)
        self.assertIn("HTTPError", webhook_events[0].response_content)

    @mock.patch("requests.post")
    def test_unboundlocal_error(self, post_mock):
        post_mock.side_effect = UnboundLocalError()
        self.order.save()
        webhook_events = WebhookEvent.objects.all()
        self.assertEqual(len(webhook_events), 1)
        self.assertFalse(webhook_events[0].delivered)
        self.assertEqual(webhook_events[0].response_http_status, 500)
        self.assertIn("", webhook_events[0].response_content)

    def test_undelivered_event_scanner(self):
        date1 = timezone.now() - timedelta(days=1)
        date2 = timezone.now() - timedelta(days=4)
        webhook = mommy.make(
            Webhook,
            user=self.user,
            target_url="http://example.com/order_update",
            secret_key="secret",
            event_type="order-created",
            retry_countdown_config={"choice": "exponential", "kwargs": {"base": 2}},
        )
        # unsuitable for event scanner query because of  retry limit
        webhookevent1 = mommy.make(
            WebhookEvent,
            modified_date=timezone.now(),
            created_date=timezone.now(),
            request_payload={},
            retry_count=TASK_RETRY_COUNT,
            delivered=False,
            webhook=webhook,
        )
        # unsuitable for event scanner query because of MAX_RETRY_COUNT
        webhookevent2 = mommy.make(
            WebhookEvent,
            request_payload={},
            retry_count=MAX_RETRY_COUNT + 1,
            delivered=False,
            webhook=webhook,
        )
        # unsuitable for event scanner query because of delivered
        webhookevent3 = mommy.make(
            WebhookEvent,
            request_payload={},
            retry_count=TASK_RETRY_COUNT + 1,
            delivered=True,
            webhook=webhook,
        )

        # unsuitable for event scanner control
        #  if undelivered_event.created_date + timedelta(seconds=2 ** undelivered_event.retry_count) <= timezone.now():
        webhookevent4 = mommy.make(
            WebhookEvent,
            request_payload={},
            retry_count=15,
            delivered=False,
            webhook=webhook,
        )

        webhookevent5 = mommy.make(
            WebhookEvent,
            request_payload={},
            retry_count=TASK_RETRY_COUNT + 1,
            delivered=False,
            webhook=webhook,
        )
        WebhookEvent.objects.filter(id=webhookevent5.id).update(
            created_date=timezone.now()
            - timedelta(2 ** (webhookevent5.retry_count + 1))
        )

        with requests_mock.Mocker() as mock:
            mock.register_uri(
                "POST",
                "http://example.com/order_update",
                [
                    {"text": "Order created", "status_code": 200},
                ],
            )
            undelivered_event_scanner()
        webhookevent1.refresh_from_db()
        webhookevent2.refresh_from_db()
        webhookevent3.refresh_from_db()
        webhookevent4.refresh_from_db()
        webhookevent5.refresh_from_db()

        # only webhookevent5 send request to given mock address
        self.assertEqual(mock.call_count, 1)

        self.assertEqual(webhookevent1.retry_count, TASK_RETRY_COUNT)
        self.assertEqual(webhookevent2.retry_count, MAX_RETRY_COUNT + 1)
        self.assertEqual(webhookevent3.retry_count, TASK_RETRY_COUNT + 1)
        self.assertEqual(webhookevent4.retry_count, 15)
        self.assertEqual(webhookevent5.retry_count, TASK_RETRY_COUNT + 2)

        self.assertEqual(webhookevent1.delivered, False)
        self.assertEqual(webhookevent2.delivered, False)
        self.assertEqual(webhookevent3.delivered, True)
        self.assertEqual(webhookevent4.delivered, False)
        self.assertEqual(webhookevent5.delivered, True)

    @mock.patch.object(LockTask, "is_exists_cache_key")
    def test_deliver_event_task(self, mock_method):
        webhookevent1 = mommy.make(
            WebhookEvent,
            modified_date=timezone.now(),
            created_date=timezone.now(),
            request_payload={},
            retry_count=0,
            delivered=False,
            uuid=uuid.uuid4(),
            webhook=self.webhook,
        )
        with requests_mock.Mocker() as mock:
            mock.register_uri(
                "POST", self.target_url, [{"text": "Bad request", "status_code": 400}]
            )
            mock_method.return_value = False
            deliver_event_task.delay(
                hook_id=webhookevent1.webhook_id,
                event_type=self.webhook.event_type,
                event_uuid=webhookevent1.uuid,
            )
        webhookevent1.refresh_from_db()
        self.assertEqual(webhookevent1.retry_count, TASK_RETRY_COUNT + 1)
        self.assertFalse(webhookevent1.delivered)

        # test new event
        with requests_mock.Mocker() as mock:
            mock.register_uri(
                "POST", self.target_url, [{"text": "Bad request", "status_code": 400}]
            )
            self.order.save()

        webhookevent2 = WebhookEvent.objects.last()
        self.assertEqual(webhookevent2.retry_count, TASK_RETRY_COUNT + 1)
        self.assertFalse(webhookevent2.delivered)
        self.assertNotEqual(webhookevent2.id, webhookevent1.id)

        WebhookEvent.objects.filter(
            id__in=[
                webhookevent1.id,
                webhookevent2.id,
            ]
        ).update(created_date=timezone.now() - timedelta(seconds=2 ** 12))

        with requests_mock.Mocker() as mock:
            mock.register_uri(
                "POST",
                self.target_url,
                [{"text": "Order event received", "status_code": 200}],
            )
            WebhookEvent.objects.filter(
                id__in=[webhookevent1.id, webhookevent2.id]
            ).update(modified_date=webhookevent1.modified_date - timedelta(days=1))
            undelivered_event_scanner()

        webhookevent1.refresh_from_db()
        webhookevent2.refresh_from_db()
        self.assertEqual(webhookevent1.retry_count, TASK_RETRY_COUNT + 2)
        self.assertEqual(webhookevent2.retry_count, TASK_RETRY_COUNT + 2)
        self.assertTrue(webhookevent1.delivered)
        self.assertTrue(webhookevent2.delivered)
        self.assertEqual(webhookevent1.response_http_status, 200)
        self.assertEqual(webhookevent2.response_http_status, 200)

    @requests_mock.mock()
    @mock.patch.object(LockTask, "is_exists_cache_key")
    def test_auth_config(self, m, mock_method):
        mock_method.return_value = False
        self.order.save()
        self.order.number = "123"
        m.register_uri(
            "POST", self.target_url2, text=self._auth_test_callback, status_code=200
        )
        self.order.save()
        event = WebhookEvent.objects.filter(webhook=self.webhook2).last()
        self.assertEqual(event.response_http_status, 200)
        self.assertTrue(event.delivered)

        self.webhook2.config = {}
        self.webhook2.save()

        self.order.number = "123123"
        self.order.save()
        event = WebhookEvent.objects.filter(webhook=self.webhook2).last()
        self.assertEqual(event.response_http_status, 401)
        self.assertFalse(event.delivered)

    def _auth_test_callback(self, request, context):
        if "Authorization" not in request.headers:
            context.status_code = 401
        return ""


def dummy_whisperer_event_callback(response, event_type, instance, payload):
    """this function creates customer in test database"""
    mommy.make(Customer)


@override_settings(CELERY_ALWAYS_EAGER=True)
class WhispererEventCallbackTestCase(TestCase):
    def setUp(self):
        self.target_url = "http://example.com/foo_bar"
        user = mommy.make(User, username="test_user")
        mommy.make(
            Webhook,
            user=user,
            target_url=self.target_url,
            secret_key="secret",
            event_type="order-created",
            retry_countdown_config={"choice": "exponential", "kwargs": {"base": 2}},
            callback="whisperer.tests.tests.dummy_whisperer_event_callback",
        )
        customer = mommy.make(Customer)
        address = mommy.make(Address)
        self.order = mommy.prepare(
            Order,
            customer=customer,
            address=address,
            number="1",
            amount="1",
            discount_amount="1",
            shipping_amount="1",
        )

    def tearDown(self):
        Customer.objects.all().delete()

    def test_runs_callback(self):
        with requests_mock.Mocker() as mock:
            mock.register_uri(
                "POST", self.target_url, text="Request Processed", status_code=200
            )

            # before callback we have one customer
            self.assertEqual(Customer.objects.count(), 1)
            self.order.save()
            webhook_events = WebhookEvent.objects.all()
            self.assertEqual(len(webhook_events), 1)
            self.assertTrue(webhook_events[0].delivered)
            event = webhook_events.first()
            self.assertEqual(self.order.pk, event.object_id)
            self.assertEqual(
                event.content_type,
                ContentType.objects.get_for_model(self.order._meta.model),
            )
            self.assertEqual(event.content_object, self.order)

            # check callback has run
            self.assertEqual(Customer.objects.count(), 2)


class RetryCountdownTestCase(TestCase):
    def test_fixed_retry_countdown(self):
        seconds = 2 * 60
        countdown = FixedRetryCountdown(seconds=seconds)
        for _ in range(100):
            retry_count = random.randint(1, 100)
            value = countdown.get_value(retry_count)
            self.assertEqual(value, seconds)

    def test_random_retry_countdown(self):
        min_value = 1 * 60
        max_value = 10 * 60
        countdown = RandomRetryCountdown(min_value=min_value, max_value=max_value)
        for _ in range(100):
            retry_count = random.randint(1, 100)
            value = countdown.get_value(retry_count)
            self.assertTrue(min_value <= value <= max_value)

    def test_linear_retry_countdown(self):
        base = 1 * 60
        countdown = LinearRetryCountdown(base=base)
        for _ in range(100):
            retry_count = random.randint(1, 100)
            value = countdown.get_value(retry_count)
            self.assertEqual(value, base * retry_count)

        limit = 10 * 60
        countdown = LinearRetryCountdown(base=base, limit=limit)
        for _ in range(100):
            retry_count = random.randint(1, 100)
            value = countdown.get_value(retry_count)
            self.assertEqual(value, min(retry_count * base, limit))

    def test_exponential_retry_countdown(self):
        base = 2
        factor = 16
        countdown = ExponentialRetryCountdown(base=base, factor=factor)
        for _ in range(15):
            retry_count = random.randint(1, 12)
            value = countdown.get_value(retry_count)
            self.assertEqual(value, (base ** retry_count) * factor)

        limit = 20 * 60
        countdown = ExponentialRetryCountdown(base=base, factor=factor, limit=limit)
        for _ in range(15):
            retry_count = random.randint(1, 12)
            value = countdown.get_value(retry_count)
            self.assertEqual(value, min((base ** retry_count) * factor, limit))



from rest_framework import response, status, viewsets

from whisperer.exceptions import WebhookAlreadyRegistered
from whisperer.models import Webhook, WebhookEvent
from whisperer.resources.filters import WebhookEventFilter, WebhookFilter
from whisperer.resources.serializers import WebhookEventSerializer, WebhookSerializer
from whisperer.services import WebhookService


class WebhookViewSet(viewsets.ModelViewSet):
    queryset = Webhook.objects.all()
    serializer_class = WebhookSerializer
    service = WebhookService()
    filter_class = WebhookFilter

    def get_queryset(self):
        queryset = super(WebhookViewSet, self).get_queryset()
        return queryset.filter(user=self.request.user)

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        try:
            self.perform_create(serializer)
        except WebhookAlreadyRegistered as exception:
            return response.Response(
                data=exception.code, status=status.HTTP_406_NOT_ACCEPTABLE
            )

        headers = self.get_success_headers(serializer.data)
        return response.Response(
            serializer.data, status=status.HTTP_201_CREATED, headers=headers
        )

    def perform_create(self, serializer):
        user = self.request.user
        serializer.instance = self.service.register_webhook(
            user, **serializer.validated_data
        )

    def perform_update(self, serializer):
        user = self.request.user
        self.service.update_webhook(
            serializer.instance, user=user, **serializer.validated_data
        )

    def perform_destroy(self, instance):
        self.service.delete_webhook(instance)


class WebhookEventViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = WebhookEvent.objects.all()
    filter_class = WebhookEventFilter
    serializer_class = WebhookEventSerializer

    def get_queryset(self):
        queryset = super(WebhookEventViewSet, self).get_queryset()
        return queryset.filter(webhook__user=self.request.user)
